// Graphic.cpp : AvP[Vp̃Gg |Cg̒`
//

#include "stdafx.h"

#include "SDIB32GL.H"
#include "util.h"
#include <cmath>
#include <list>
using namespace std;

/*** C^[tFCX̏ ***/
LPSDIB32GLFACTORY	lpSDFactory = NULL;
LPSDIB32GL			lpBack		= NULL;
LPSDIB32GL			lpBg		= NULL;

#define T_VERTEX				3		// _
#define T_SHADOW_TIMING			1		// gCAOVhE^C 
#define T_ROTATE_SPEED			6		// ]Xs[h
#define T_START_ALPHA			220		// Jnl
#define T_SCALE_SPEED			0.03f	// XP[Xs[h
#define T_START_SIZE			300		// JnTCY
#define TS_ALPHA_SPEED			6		// Xs[h
#define TC_TRIANGLE_TIMING		15		// gCAO^C

#define SINF(a)	sinftbl[a&511]
#define COSF(a) sinftbl[(a+128)&511]
float sinftbl[512];

// e^XNx[X
class CTaskBase {
	friend class CTaskManager;
protected:
	bool	m_bKill;
public:
	CTaskBase() : m_bKill(false) {}
	virtual ~CTaskBase() {}
	virtual void Task( void* param ) = 0;
};

// e^XN}l[W[
class CTaskManager {
protected:
	typedef list<CTaskBase*> CTSKLIST;
	typedef list<CTaskBase*>::iterator CTSKLISTIT;

	CTSKLIST m_TaskList;

public:
	CTaskManager() {}
	virtual ~CTaskManager() { KillAllTask(); }

	void CallTask( void ) {
		CTSKLISTIT it = m_TaskList.begin();
		while( it != m_TaskList.end() ) {
			(*it)->Task( this );
			if( (*it)->m_bKill ) {
				delete *it;
				it = m_TaskList.erase( it );
				continue;
			}
			it++;
		}
	}

	void AddTask( CTaskBase* pTskBase ) {
		m_TaskList.push_back( pTskBase );
	}

	void KillAllTask( void ) {
		CTSKLISTIT it = m_TaskList.begin();
		while( it != m_TaskList.end() ) {
			delete *it;
			it = m_TaskList.erase( it );
		}
	}
};

class CTriangle;

// gCAOVhE^XNNX
class CTShadow : public CTaskBase
{
protected:
	POINT			m_vtx[ T_VERTEX ];
	long			m_alpha;
	SDIB32GLCOLOR	m_color;
public:
	CTShadow( LPPOINT ppos, long alpha, SDIB32GLCOLOR color ) : m_alpha(alpha), m_color(color){
		memmove( m_vtx, ppos, sizeof(POINT)*T_VERTEX );
	}
	virtual ~CTShadow() {}
	
	virtual void Task( void* param ) {
		m_alpha -= TS_ALPHA_SPEED;

		if( m_alpha < 0 ) {
			m_bKill = true;
			return;
		}

		SDIB32GLARGB Argb;
		Argb.Color = m_color;
		Argb.Alpha = m_alpha;

		// ZC̕`
		for( int i=0; i<T_VERTEX; i++) {
			lpBack->LineAdd( m_vtx[i].x, m_vtx[i].y,
				m_vtx[(i+1)%T_VERTEX].x, m_vtx[(i+1)%T_VERTEX].y, Argb.Color);
		}
	}
};

#define T_MAX_COLOR_TBL	8

// gCAO^XNNX
class CTriangle : public CTaskBase
{
protected:
	POINT			m_vtx[ T_VERTEX ];
	POINT			m_def;
	long			m_alpha;
	long			m_angle;
	long			m_counter;
	SDIB32GLCOLOR	m_color;
	float			m_scale;
	static SDIB32GLCOLOR m_coltbl[T_MAX_COLOR_TBL];

public:
	CTriangle() : m_alpha(T_START_ALPHA), m_angle(0), m_counter(0), m_scale(1.f) {
		for( int i=0; i<T_VERTEX; i++ ) {
			m_vtx[i].x = COSF(i*(512/T_VERTEX))*T_START_SIZE;
			m_vtx[i].y = SINF(i*(512/T_VERTEX))*T_START_SIZE;
		}
		m_def.x = rand()%CLIENT_WIDTH;
		m_def.y = rand()%CLIENT_HEIGHT;
		m_color = m_coltbl[rand()%T_MAX_COLOR_TBL];
	};
	virtual ~CTriangle() { }
	
	virtual void Task( void* param ) {
		POINT pos[ T_VERTEX ];

		for( int i=0; i<T_VERTEX; i++ ) {
			pos[i].x = COSF(m_angle)*m_vtx[i].x+SINF(m_angle)*m_vtx[i].y;
			pos[i].y = -SINF(m_angle)*m_vtx[i].x+COSF(m_angle)*m_vtx[i].y;
			pos[i].x *= m_scale;
			pos[i].y *= m_scale;
			pos[i].x += m_def.x;
			pos[i].y += m_def.y;
		}

		if( m_counter == 0 ) {
			static_cast<CTaskManager*>(param)->AddTask( new CTShadow( pos, m_alpha, m_color ) );
		}

		m_scale -= T_SCALE_SPEED;

		if( m_scale < 0.f ) {
			m_bKill = true;
			return;
		}

		m_angle += T_ROTATE_SPEED;
		++m_counter %= T_SHADOW_TIMING;

		SDIB32GLARGB Argb;
		Argb.Color = m_color;
		Argb.Alpha = m_alpha;

		// ZC̕`
		for( int i=0; i<T_VERTEX; i++) {
			lpBack->LineAdd( pos[i].x, pos[i].y,
				pos[(i+1)%T_VERTEX].x, pos[(i+1)%T_VERTEX].y, Argb.Color );
		}
	}
};

SDIB32GLCOLOR CTriangle::m_coltbl[T_MAX_COLOR_TBL] = { 
	SDIB32GLCOLOR_ARGB(0,255,  0,  0), SDIB32GLCOLOR_ARGB(0,255,128,  0), 
	SDIB32GLCOLOR_ARGB(0,255,  0,255), SDIB32GLCOLOR_ARGB(0,  0,255,128),
	SDIB32GLCOLOR_ARGB(0,  0,  0,255), SDIB32GLCOLOR_ARGB(0,  0,255,255),
	SDIB32GLCOLOR_ARGB(0,255,255,  0), SDIB32GLCOLOR_ARGB(0,255,  0,128),
};

// gCAO^XN^XNNX
class CTCreator : public CTaskBase {
protected:
	long m_nCounter;
public:
	CTCreator() : m_nCounter(0) {}
	virtual ~CTCreator() {}

	void Task( void* param ) {
		if( m_nCounter == 0) {
			static_cast<CTaskManager*>(param)->AddTask( new CTriangle );
		}
		++m_nCounter %= TC_TRIANGLE_TIMING;
	}
};


CTaskManager TskManager;


/*----------------------------------------------------------------------------*
|	App
*----------------------------------------------------------------------------*/
BOOL AppInitialize( void )
{
	HRESULT hr;

	try
	{
		// t@NgC^[tFCX擾
		hr = SDIB32GLFactoryCreate( 
			SDIB32GLCREATE_AUTO, (void**)&lpSDFactory, IID_ISDib32GLFactory );

		if( FAILED( hr ) )
		{
			throw exception("t@NgC^[tFCXs");
		}

		// NCAg̈TCỸobNobt@쐬
		hr = lpSDFactory->CreateBuffer(
			CLIENT_WIDTH, CLIENT_HEIGHT, 0, NULL, &lpBack );

		if( FAILED( hr ) )
		{
			throw exception("obNobt@̍쐬Ɏs");
		}

		RANDOMIZE;

		// BGǂݍ
		hr = lpSDFactory->CreateBufferFromFile(
			MEDIAPATH( OLESTR("BackGround.bmp") ), 0, NULL, 0, NULL, &lpBg );

		if( FAILED( hr ) )
		{
			throw exception("t@C̓ǂݍ݂Ɏs");
		}
	}


	// G[
	catch( exception& e )
	{
		
		MessageBox( hWnd, e.what(), szTitle, MB_OK | MB_ICONSTOP );

		// N[Abv
		AppCleanUp();

		return FALSE;
	}


	SetTextMode( lpBack );

	// TCe[u쐬
	for( int i=0; i<512; i++ )
	{
		sinftbl[i] = sinf(3.1415926535*i/256);
	}

	// 邳ƂĂ݂
	SIZE size;
	lpBg->GetSize(&size);
	lpBg->Bright( 0, 0, size.cx, size.cy, -128 );

	// NGC^[^XN̒ǉ
	TskManager.AddTask( new CTCreator );

	return TRUE;
}


/*----------------------------------------------------------------------------*
|	AppACh
*----------------------------------------------------------------------------*/
void AppIdle( void )
{

	lpBack->Blt( lpBg, 0, 0, NULL, NULL, NULL );

	// ^XNĂяo
	TskManager.CallTask();

	// ̕`
	DrawInfo( lpBack );


	// ĕ`
	//*
	HDC hDC = GetDC( hWnd );
	lpBack->Draw( hDC );
	ReleaseDC( hWnd, hDC );
	/*/
	InvalidateRect( hWnd, NULL, FALSE );
	UpdateWindow( hWnd );
	//*/
}


/*----------------------------------------------------------------------------*
|	AppN[Abv
*----------------------------------------------------------------------------*/
void AppCleanUp( void )
{
	SAFE_RELEASE( lpBg );
	SAFE_RELEASE( lpBack );
	SAFE_RELEASE( lpSDFactory );
}


/*----------------------------------------------------------------------------*
|	WM_PAINT
*----------------------------------------------------------------------------*/
void OnPaint( HDC hdc )
{
	if( lpBack == NULL ) return;

	lpBack->Draw( hdc );
}
