#ifndef _INC_COMUTIL
#define _INC_COMUTIL


// C^[tFCXGg\
typedef struct tagINTERFACE_ENTRY
{
	const IID*	pIID;
	DWORD		dwOffset;
}INTERFACE_ENTRY;

// C^[tFCXLXg֐
HRESULT __InternalQueryInterface( void* This, const INTERFACE_ENTRY* pEntries, REFIID riid, void** ppv );

// vptr̃ItZbgl擾
#define VPTR_OFFSET( Class, Interface )	\
	(DWORD(static_cast<Interface*>(reinterpret_cast<Class*>(0x10000000)))-0x10000000)

// COM}bvJn
#define _BEGIN_COM_MAP( Class )	\
protected:	\
	CComRef	m_cRef;	\
public:	\
	STDMETHODIMP QueryInterface( REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject) {	\
		return ::__InternalQueryInterface( this, _GetEntries(), riid, ppvObject );	\
	}	\
	STDMETHODIMP_(ULONG) AddRef( void ) {	\
		CComSimpleModule::LockModule();	\
		return ++m_cRef;	\
	}	\
	STDMETHODIMP_(ULONG) Release( void ) {	\
		CComSimpleModule::UnlockModule();	\
		if( --m_cRef == 0 ) { delete this; return 0; }	\
		return m_cRef;	\
	}	\
	typedef Class _ComMapClass;		\
	const INTERFACE_ENTRY* _GetEntries( void ) {	\
	static const INTERFACE_ENTRY _Entries[] = {

// COMC^[tFCXGg[
#define _COM_INTERFACE_ENTRY( Interface )	\
	{&IID_##Interface, VPTR_OFFSET(_ComMapClass, Interface) },

// COMC^[tFCXGg[
#define _COM_INTERFACE_ENTRY_AS( Iid, Interface )	\
	{&IID_##Iid, VPTR_OFFSET(_ComMapClass, Interface) },

// COM}bvI
#define _END_COM_MAP()	\
	{NULL, 0}}; return _Entries; }


typedef IClassFactory* (*FACTORY_CREATEINSTANCE)( void );

// COMIuWFNgGg\
typedef struct tagOBJECT_ENTRY
{
	const CLSID*			pClsID;
	FACTORY_CREATEINSTANCE	ClassFactory_CreateInstance;
	const char*				lpszPorgID;
	const char*				lpszComName;
}OBJECT_ENTRY;

// COMIuWFNg}bvJn
#define _BEGIN_OBJECT_MAP( Entries )	\
	static OBJECT_ENTRY Entries[] = {	\

// COMIuWFNgGg
#define _OBJECT_ENTRY(Clsid, Class, ProgID, ComName)	\
	{&Clsid, CComSimpleCreator<Class>::CreateInstance, ProgID, ComName},

// COMIuWFNg}bvI
#define _END_OBJECT_MAP()	\
{NULL, NULL, NULL, NULL} };


/*----------------------------------------------------------------------------*
|	COMt@XJE^NX
*----------------------------------------------------------------------------*/
class CComRef
{
protected:
	
	LONG	m_cRef;
public:

	CComRef( LONG cRef=1 ) : m_cRef(cRef) {}
	virtual ~CComRef(){}

	CComRef& operator ++(){ ::InterlockedIncrement(&m_cRef); return *this; }	// ++this
	CComRef  operator ++(int){ CComRef tmp=*this; ++*this; return tmp; }		// this++
	CComRef& operator --(){ ::InterlockedDecrement(&m_cRef); return *this; }	// --this
	CComRef  operator --(int){ CComRef tmp=*this; --*this; return tmp; }		// this--
	LONG operator +(int n){ return m_cRef+n; }
	LONG operator -(int n){ return m_cRef-n; }
	bool operator !(){ return !m_cRef; }
	bool operator ==(int n){ return (m_cRef == n); }
	bool operator !=(int n){ return (m_cRef != n); }
	operator LONG(){ return m_cRef; }
};


/*----------------------------------------------------------------------------*
|	COMNXt@Ngev[g
*----------------------------------------------------------------------------*/
template <class T>
class CComSimpleClassFactory : public IClassFactory
{
public:
	CComSimpleClassFactory() _THROW0() {}

	_BEGIN_COM_MAP( CComSimpleClassFactory<T> )
		_COM_INTERFACE_ENTRY( IClassFactory )
	_END_COM_MAP()

	STDMETHOD( CreateInstance )( IUnknown* pUnkOuter, REFIID riid, void **ppv ) {
		if( ppv == NULL ) return E_POINTER;
		*ppv = NULL;

		if( pUnkOuter != NULL ) return CLASS_E_NOAGGREGATION;

		T* pClassObject = new(nothrow) T;

		if( pClassObject == NULL ) return E_OUTOFMEMORY;

		HRESULT hr = pClassObject->QueryInterface( riid, ppv );

		pClassObject->Release();
		return hr;
	}

	STDMETHOD( LockServer )( BOOL fLock ) {
		if( fLock ) {
			CComSimpleModule::LockModule();
		}else {
			CComSimpleModule::UnlockModule();
		}
		return S_OK;
	}
};


/*----------------------------------------------------------------------------*
|	COM쐬NX
*----------------------------------------------------------------------------*/
template <class T>
class CComSimpleCreator {
public:
	static IClassFactory* CreateInstance( void ) {
		return new(nothrow) CComSimpleClassFactory<T>;
	}
};


/*----------------------------------------------------------------------------*
|	COMW[NX
*----------------------------------------------------------------------------*/
class CComSimpleModule
{
protected:
	
	static char				m_szModulePath[MAX_PATH];
	static OBJECT_ENTRY*	m_pObjectMap;
	static CComRef			m_cObjectsAndLocks;

protected:

	static HRESULT GetAnsiStringCLSID( REFCLSID rclsid, char* lpszCLDID, long nCharMax );
	static HRESULT RegisterClassKey( REFCLSID rclsid, LPCSTR lpszModulePath, LPCSTR lpszComName, LPCSTR lpszProgID );
	static HRESULT UnregisterClassKey( REFCLSID rclsid, LPCSTR lpszProgID );

public:

	CComSimpleModule() {};
	virtual ~CComSimpleModule() {};

	/*  */
	static void Initialize( HMODULE hModule, OBJECT_ENTRY* pObjectMap );

	/* W[bNJE^ */
	static void LockModule( void ){ m_cObjectsAndLocks++; }
	static void UnlockModule( void ){ m_cObjectsAndLocks--; }

	/* DLLɃGLX|[g֐ */
	static HRESULT GetClassObject( REFCLSID rclsid, REFIID riid, void** ppv );
	static HRESULT RegisterServer( void );
	static HRESULT UnregisterServer( void );
	static LONG GetLockCount( void ){ return m_cObjectsAndLocks; }
	
};


#endif /* _INC_COMUTIL */